# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# * The software is provided "as is" without any warranty, neither express
#   nor implied.
# * In no event will the author be held liable for any damage(s) arising
#   from the use of the software.
# * Redistribution of the software is allowed only in unmodified form.
# * Permission is granted to use the software solely for personal private
#   and non-commercial purposes.
# * An individuals use of the software in his or her capacity or function
#   as an agent, (independent) contractor, employee, member or officer of
#   a business, corporation or organization (commercial or non-commercial)
#   does not qualify as personal private and non-commercial purpose.
# * Without written approval from the author the software must not be used
#   for a business, for commercial, corporate, governmental, military or
#   organizational purposes of any kind, or in a commercial, corporate,
#   governmental, military or organizational environment of any kind.

if [ $# -eq 1 -a "$1" == 'amd64' ]
then
as --64 --no-pad-sections --noexecstack -o memory.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# Unix System V calling convention for AMD64 platform:
# - first 6 integer or pointer arguments (from left to right) are passed
#   in registers RDI/R7, RSI/R6, RDX/R2, RCX/R1, R8 and R9
#   (R10 is used as static chain pointer in case of nested functions);
# - surplus arguments are pushed on stack in reverse order (from right to
#   left), 8-byte aligned;
# - 128-bit integer arguments are passed as pair of 64-bit integer arguments,
#   low part before/below high part;
# - 128-bit integer result is returned in registers RAX/R0 (low part) and
#   RDX/R2 (high part);
# - 64-bit integer or pointer result is returned in register RAX/R0;
# - 32-bit integer result is returned in register EAX;
# - registers RBX/R3, RSP/R4, RBP/R5, R12, R13, R14 and R15 must be
#   preserved;
# - registers RAX/R0, RCX/R1, RDX/R2, RSI/R6, RDI/R7, R8, R9, R10 (in
#   case of normal functions) and R11 are volatile and can be clobbered;
# - stack is 16-byte aligned: callee must decrement RSP by 8+n*16 bytes
#   before calling other functions (CALL instruction pushes 8 bytes);
# - a "red zone" of 128 bytes below the stack pointer can be clobbered.

.file	"memory.s"
.arch	generic64
.code64
.ident	"Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of buffer
					# sil = character
					# rdx = count
memchr:
	xor	eax, eax		# rax = 0
	test	rdx, rdx
	jz	.Lnull			# count = 0?

	mov	r8, 0x0101010101010101
.if 0
	mov	r9, 0x8080808080808080
.elseif 0
	imul	r9, r8, 128		# r9 = 0x8080808080808080
.else
	mov	r9, r8
	ror	r9, 1			# r9 = 0x8080808080808080
.endif
	movzx	esi, sil		# rsi = character
	imul	rsi, r8			# rsi = character
					#     | character << 8
					#     | character << 16
					#     | character << 24
					#     | character << 32
					#     | character << 40
					#     | character << 48
					#     | character << 56
	add	rdx, rdi		# rdx = address after buffer
	mov	ecx, edi
	and	ecx, 7			# rcx = address of buffer % 8
					#     = 8 - number of unaligned bytes
	jz	.Laligned		# address of buffer % 8 = 0?
.Lunaligned:
	sub	rdi, rcx		# rdi = aligned address before buffer
	shl	ecx, 3			# rcx = (8 - number of unaligned bytes) * 8
					#     = 64 - number of unaligned bits
	not	rax			# rax = ~0
	shl	rax, cl			# rax = ~0 for unaligned bytes, 0 elsewhere
	not	rax			# rax = 0 for unaligned bytes, ~0 elsewhere
	mov	rcx, [rdi]		# rcx = unaligned bytes
	xor	rcx, rsi		# rcx = unaligned bytes ^ multiplied character
	or	rcx, rax		# rcx = '\0' for matching unaligned bytes
	jmp	.Ltwiddle
.Lnext:
	add	rdi, 8			# rdi = address of next 8 aligned bytes
	cmp	rdi, rdx
	jae	.Lnull			# address after buffer?
.Laligned:
	mov	rcx, [rdi]		# rcx = next 8 aligned bytes
	xor	rcx, rsi		# rcx = next 8 aligned bytes ^ multiplied character
					#     = '\0' for matching bytes
.Ltwiddle:
	mov	rax, rcx
	sub	rcx, r8
	not	rax
	and	rcx, r9
	and	rax, rcx		# rax = '\200' for matching bytes, '\0' elsewhere
	jz	.Lnext			# no match in any byte?
.Lmatch:
	bsf	rax, rax		# rax = offset of matching byte * 8 + 7
					#     = {7, 15, 23, 31, 39, 47, 55, 63}
	shr	eax, 3			# rax = offset of matching byte
					#     = {0, 1, 2, 3, 4, 5, 6, 7}
	add	rax, rdi		# rax = address of matching byte
.if 0
	cmp	rax, rdx		# CF = (address inside buffer)
	sbb	rdx, rdx		# rdx = (address inside buffer) ? -1 : 0
	and	rax, rdx		# rax = address of character
.else
	xor	ecx, ecx		# rcx = 0
	cmp	rax, rdx		# CF = (address inside buffer)
	cmovnb	rax, rcx		# rax = address of character
.endif
.Lnull:
	ret

.size	memchr, .-memchr
.type	memchr, @function
.global	memchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of buffer
					# sil = character
					# rdx = count
memchr:
	test	rdx, rdx
	jz	.Lnone			# count = 0?

	mov	al, sil			# al = character
.Lloop:
	scasb				# rdi = ++buffer
	je	.Lmatch

	dec	rdx			# rdx = --count
	jnz	.Lloop			# count <> 0?
.Lnone:
	xor	eax, eax
	ret
.Lmatch:
	lea	rax, [rdi-1]		# rax = address of character
	ret

.size	memchr, .-memchr
.type	memchr, @function
.global	memchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of buffer
					# sil = character
					# rdx = count
memchr:
	mov	rcx, rdx		# rcx = count
	mov	eax, esi		# rax = character
	repne	scasb
.if 0
	lea	rax, [rdi-1]		# rax = address of character
	test	rcx, rcx
	cmovz	rax, rcx		# rax = (rcx = 0) ? 0 : address of character
.else
	dec	rdi
	neg	rcx			# CF = (ecx <> 0)
					#    = ([rdi] = character)
	sbb	rax, rax		# rax = (rcx = 0) ? 0 : -1
	and	rax, rdi		# rax = (rcx = 0) ? 0 : address of character
.endif
	ret

.size	memchr, .-memchr
.type	memchr, @function
.global	memchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memcmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memcmp.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of left buffer
					# rsi = address of right buffer
					# rdx = count
memcmp:
#	cmp	rdi, rsi
#	je	.Lequal			# address of left buffer = address of right buffer?

	test	rdx, rdx
	jz	.Lequal			# count = 0?
.Lloop:
	lodsb				# al = *right++
	scasb
	jne	.Lnotequal		# al = *left++?

	dec	rdx			# rdx = --count
	jnz	.Lloop			# count <> 0?
.Lequal:
	xor	eax, eax
	ret
.Lnotequal:
	sbb	eax, eax		# eax = (*left > *right) ? -1 : 0
	or	eax, 1			# eax = (*left > *right) ? -1
					#     : (*left < *right) ? 1 : 0
					#     = {-1, 0, 1}
	neg	eax
	ret

.size	memcmp, .-memcmp
.type	memcmp, @function
.global	memcmp
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memcmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memcmp.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of left buffer
					# rsi = address of right buffer
					# rdx = count
memcmp:
#	xor	eax, eax		# rax = 0
#	cmp	rdi, rsi
#	je	.Lequal			# address of left buffer = address of right buffer?

	mov	rcx, rdx		# rcx = count
#	jrcxz	.Lequal			# count = 0?

	xor	eax, eax		# rax = 0,
					# CF = 0, ZF = 1 (required when count is 0)
	repe	cmpsb
	seta	al			# rax = (*right > *left) ? 1 : 0
	sbb	rax, 0			# rax = (*right < *left) ? -1
					#     : (*right > *left) ? 1 : 0
					#     = {1, 0, -1}
	neg	rax
.Lequal:
	ret

.size	memcmp, .-memcmp
.type	memcmp, @function
.global	memcmp
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memcpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memcpy.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
					# rdx = count
memcpy:
	mov	rcx, rdi		# rcx = address of destination
	test	rdx, rdx
	jz	.Lnone			# count = 0?
.Lloop:
	lodsb				# al = *source++
	stosb				# *destination++ = al
	dec	rdx			# rdx = --count
	jnz	.Lloop			# count <> 0?
.Lnone:
	mov	rax, rcx		# rcx = address of destination
	ret

.size	memcpy, .-memcpy
.type	memcpy, @function
.global	memcpy
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memcpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memcpy.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
					# rdx = count
memcpy:
	mov	rax, rdi		# rax = address of destination
	mov	rcx, rdx		# rcx = count
	rep	movsb
	ret

.size	memcpy, .-memcpy
.type	memcpy, @function
.global	memcpy
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memmove.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memmove.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of destination
					# rsi = address of source
					# rdx = count
memmove:
	mov	rax, rdi		# rax = address of destination
	mov	rcx, rdx		# rcx = count
#	jrcxz	.Lnone			# count = 0?

	cmp	rdi, rsi
	jb	.Ldefault		# address of destination < address of source?
#	je	.Lnone			# address of destination = address of source?

	mov	rdx, rdi
	sub	rdx, rsi
	jb	.Loverlap		# address of destination - address of source < count?
.Ldefault:
	rep	movsb
.Lnone:
	ret
.Loverlap:
	lea	rdi, [rdi+rcx-1]	# end of destination
	lea	rsi, [rsi+rcx+1]	# end of source
	std
	rep	movsb
	cld
	ret

.size	memmove, .-memmove
.type	memmove, @function
.global	memmove
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memrchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of buffer
					# sil = character
					# rdx = count
memrchr:
	xor	eax, eax		# rax = 0
	test	rdx, rdx
	jz	.Lnull			# count = 0?

	mov	r8, 0x0101010101010101
.if 0
	mov	r9, 0x8080808080808080
.elseif 0
	imul	r9, r8, 128		# r9 = 0x8080808080808080
.else
	mov	r9, r8
	ror	r9, 1			# r9 = 0x8080808080808080
.endif
	movzx	esi, sil		# rsi = character
	imul	rsi, r8			# rsi = character
					#     | character << 8
					#     | character << 16
					#     | character << 24
					#     | character << 32
					#     | character << 40
					#     | character << 48
					#     | character << 56
	add	rdx, rdi		# rdx = address after buffer
	mov	ecx, edx
	and	ecx, 7			# rcx = address after buffer % 8
					#     = 8 - number of tail bytes
	jz	.Laligned		# address after buffer % 8 = 0?
.Lunaligned:
	sub	rdx, rcx		# rdx = aligned address of tail bytes
	shl	ecx, 3			# rcx = (8 - number of unaligned bytes) * 8
					#     = 64 - number of unaligned bits
	not	rax			# rax = ~0
	shl	rax, cl			# rax = '\0' for tail bytes, ~0 elsewhere
	mov	rcx, [rdx]		# rcx = tail bytes
	xor	rcx, rsi		# rcx = tail bytes ^ multiplied character
	or	rcx, rax		# rcx = '\0' for matching tail bytes
	jmp	.Ltwiddle
.Lprevious:
	sub	rdx, 8			# rdx = address of previous 8 aligned bytes
	cmp	rdx, rdi
	jb	.Lnull			# address before buffer?
.Laligned:
	mov	rcx, [rdx]		# rcx = previous 8 aligned bytes
	xor	rcx, rsi		# rcx = previous 8 aligned bytes ^ multiplied character
					#     = '\0' for matching bytes
.Ltwiddle:
	mov	rax, rcx
	sub	rcx, r8
	not	rax
	and	rcx, r9
	and	rax, rcx		# rax = '\200' for matching bytes, '\0' elsewhere
	jz	.Lprevious		# no match in any byte?
.Lmatch:
	bsr	rax, rax		# rax = offset of matching byte * 8 + 7
					#     = {63, 55, 47, 39, 31, 23, 15, 7}
	shr	eax, 3			# rax = offset of matching byte
					#     = {7, 6, 5, 4, 3, 2, 1, 0}
	add	rax, rdx		# rax = address of matching byte
.if 0
	cmp	rax, rdi		# CF = (address of matching byte < address of buffer)
	cmc				# CF = (address of matching byte >= address of buffer)
	sbb	rdx, rdx		# rdx = (address inside buffer) ? -1 : 0
	and	rax, rdx		# rax = address of character
.else
	xor	edx, edx
	cmp	rax, rdi		# CF = (address of character < address of buffer)
	cmovb	rax, rdx		# rax = address of character
.endif
.Lnull:
	ret

.size	memrchr, .-memrchr
.type	memrchr, @function
.global	memrchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memrchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of buffer
					# sil = character
					# rdx = count
memrchr:
	test	rdx, rdx
	jz	.Lnone			# count = 0?

	mov	al, sil			# al = character
	lea	rdi, [rdi+rdx-1]	# rdi = end of buffer
	std
.Lloop:
	scasb				# rdi = --buffer
	je	.Lmatch

	dec	rdx			# rdx = --count
	jnz	.Lloop			# count <> 0?

	cld
.Lnone:
	xor	eax, eax
	ret
.Lmatch:
	cld
	lea	rax, [rdi+1]		# rax = address of character
	ret

.size	memrchr, .-memrchr
.type	memrchr, @function
.global	memrchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memrchr.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of buffer
					# sil = character
					# rdx = count
memrchr:
	mov	rcx, rdx		# rcx = count
	mov	eax, esi		# rax = character
	lea	rdi, [rdi+rdx-1]	# rdi = end of buffer
	std
	repne	scasb
	cld
.if 0
	lea	rax, [rdi+1]		# rax = address of character
	test	rcx, rcx
	cmovz	rax, rcx		# rax = (rcx = 0) ? 0 : address of character
.else
	inc	rdi
	neg	rcx			# CF = (ecx <> 0)
					#    = ([rdi] = character)
	sbb	rax, rax		# rax = (rcx = 0) ? 0 : -1
	and	rax, rdi		# rax = (rcx = 0) ? 0 : address of character
.endif
	ret

.size	memrchr, .-memrchr
.type	memrchr, @function
.global	memrchr
.end
+EOF

as --64 --no-pad-sections --noexecstack -o memset.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memset.s"
.arch	generic64
.code64
.intel_syntax noprefix
.text
					# rdi = address of buffer
					# sil = character
					# rdx = count
memset:
	mov	rcx, rdx		# rcx = count
	mov	eax, esi		# rax = character
	mov	rdx, rdi		# rdx = address of buffer
	repne	stosb
	mov	rax, rdx		# rax = address of buffer
	ret

.size	memset, .-memset
.type	memset, @function
.global	memset
.end
+EOF

ar rsv memory-amd64.a memory.o memchr.o memcmp.o memcpy.o memmove.o memrchr.o memset.o
rm memory.o memchr.o memcmp.o memcpy.o memmove.o memrchr.o memset.o
elif [ $# -eq 1 -a "$1" = 'i386' ]
then
as --32 --no-pad-sections --noexecstack -o memory.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# "cdecl" calling and naming convention for i386 platform:
# - arguments are pushed on stack in reverse order (from right to left),
#   4-byte aligned;
# - 64-bit integer arguments are passed as pair of 32-bit integer arguments,
#   low part below high part;
# - 64-bit integer result is returned in registers EAX (low part) and
#   EDX (high part);
# - 32-bit integer or pointer result is returned in register EAX;
# - registers EAX, ECX and EDX are volatile and can be clobbered;
# - registers EBX, ESP, EBP, ESI and EDI must be preserved.

.file	"memory.s"
.arch	generic32
.code32
.ident	"Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memchr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = character
					# [esp+4] = address of buffer
memchr:
	xor	eax, eax		# eax = 0
	cmp	eax, [esp+12]
	je	.Lreturn		# count = 0?

	cdq				# edx = 0
	mov	dl, [esp+8]		# edx = character
	imul	edx, 0x01010101		# edx = character
					#     | character << 8
					#     | character << 16
					#     | character << 24
	mov	ecx, [esp+4]		# ecx = address of buffer
	add	[esp+12], ecx		# count = address after buffer
	push	ebx
	mov	ebx, ecx		# ebx = address of buffer
	and	ecx, 3			# ecx = address of buffer % 4
					#     = 4 - number of unaligned bytes
	jz	.Laligned		# address of buffer % 4 = 0?
.Lunaligned:
	sub	ebx, ecx		# ebx = aligned address before buffer
	shl	ecx, 3			# ecx = (4 - number of unaligned bytes) * 8
					#     = 32 - number of unaligned bits
	dec	eax			# eax = ~0
	shl	eax, cl			# eax = ~0 for unaligned bytes, 0 elsewhere
	not	eax			# eax = 0 for unaligned bytes, ~0 elsewhere
	mov	ecx, [ebx]		# ecx = unaligned bytes
	xor	ecx, edx		# ecx = unaligned bytes ^ multiplied character
	or	eax, ecx		# eax = '\0' for matching bytes
	jmp	.Ltwiddle
.Lnext:
	add	ebx, 4			# ebx = address of next 4 aligned bytes
	cmp	ebx, [esp+16]
	jae	.Lnull			# address after buffer?
.Laligned:
	mov	eax, [ebx]		# eax = next 4 aligned bytes
	xor	eax, edx		# eax = next 4 aligned bytes ^ multiplied character
					#     = '\0' for matching bytes
.Ltwiddle:
	mov	ecx, eax
	sub	eax, 0x01010101
	not	ecx
	and	eax, 0x80808080
	and	eax, ecx		# eax = '\200' for matching bytes, '\0' elsewhere
	jz	.Lnext			# no match in any byte?
.Lmatch:
	bsf	eax, eax		# eax = offset of matching byte * 8 + 7
					#     = {7, 15, 23, 31}
	shr	eax, 3			# eax = offset of matching byte
					#     = {0, 1, 2, 3}
	add	eax, ebx		# eax = address of matching byte
	cmp	eax, [esp+16]		# CF = (address inside buffer)
	sbb	ecx, ecx		# ecx = (address inside buffer) ? -1 : 0
	and	eax, ecx		# eax = address of character
.Lnull:
	pop	ebx
.Lreturn:
	ret

.size	memchr, .-memchr
.type	memchr, @function
.global	memchr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memchr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = character
					# [esp+4] = address of buffer
memchr:
	mov	ecx, [esp+12]		# ecx = count
	test	ecx, ecx
	jz	.Lnone			# count = 0?

	mov	edx, [esp+8]		# dl = character
	mov	eax, [esp+4]		# eax = address of buffer
.Lloop:
	cmp	dl, [eax]
	je	.Lmatch			# dl = *buffer?

	inc	eax			# eax = ++buffer
	dec	ecx			# ecx = --count
	jnz	.Lloop			# count <> 0?
.Lnone:
	mov	eax, ecx
.Lmatch:
	ret

.size	memchr, .-memchr
.type	memchr, @function
.global	memchr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memchr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = character
					# [esp+4] = address of buffer
memchr:
	mov	edx, edi
	mov	edi, [esp+4]		# edi = address of buffer
	mov	eax, [esp+8]		# eax = character
	mov	ecx, [esp+12]		# ecx = count
	repne	scasb
	dec	edi			# edi = address of character
	neg	ecx			# CF = (ecx <> 0)
					#    = ([edi] = character)
	sbb	eax, eax		# eax = (ecx = 0) ? 0 : -1
	and	eax, edi		# eax = (ecx = 0) ? 0 : address of character
	mov	edi, edx
	ret

.size	memchr, .-memchr
.type	memchr, @function
.global	memchr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memcmp.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memcmp.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = address of right buffer
					# [esp+4] = address of left buffer
memcmp:
	mov	eax, [esp+4]		# eax = address of left buffer
	mov	edx, [esp+8]		# edx = address of right buffer
	cmp	edx, eax
	je	.Lequal			# address of left buffer = address of right buffer?
.if 0
	mov	ecx, [esp+12]		# ecx = count
	jecxz	.Lequal			# count = 0?
.else
	cmp	ebx, ebx		# CF = 0, ZF = 1 (required when count is 0)
	mov	ecx, [esp+12]		# ecx = count
.endif
	xchg	esi, eax		# esi = address of left buffer
	xchg	edi, edx		# edi = address of right buffer
	repe	cmpsb
	mov	edi, edx
	mov	esi, eax
	seta	al
	movzx	eax, al			# eax = (left buffer > right buffer) ? 1 : 0
	sbb	eax, 0			# eax = (left buffer < right buffer) ? -1
					#     : (left buffer > right buffer) ? 1 : 0
	ret
.Lequal:
	xor	eax, eax
	ret

.size	memcmp, .-memcmp
.type	memcmp, @function
.global	memcmp
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memcpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memcpy.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = address of source
					# [esp+4] = address of destination
memcpy:
	mov	eax, [esp+4]		# eax = address of destination
	mov	edx, [esp+8]		# edx = address of source
	cmp	edx, eax
	je	.Lnone			# address of source = address of destination?

	mov	ecx, [esp+12]		# ecx = count
	test	ecx, ecx
	jz	.Lnone			# count = 0?

	push	ebx
.Lloop:
	dec	ecx
	mov	bl, [edx+ecx]
	mov	[eax+ecx], bl
	jnz	.Lloop			# count <> 0?

	pop	ebx
.Lnone:
	ret

.size	memcpy, .-memcpy
.type	memcpy, @function
.global	memcpy
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memcpy.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memcpy.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = address of source
					# [esp+4] = address of destination
memcpy:
	mov	eax, [esp+4]		# eax = address of destination
	mov	edx, [esp+8]		# edx = address of source
	cmp	edx, eax
	je	.Lnone			# address of source = address of destination?

	mov	ecx, [esp+12]		# ecx = count
#	jecxz	.Lnone			# count = 0?

	xchg	esi, edx		# esi = address of source
	xchg	edi, eax		# edi = address of destination
	rep	movsb
	mov	esi, edx
	mov	edi, eax
	mov	eax, [esp+4]		# eax = address of destination
.Lnone:
	ret

.size	memcpy, .-memcpy
.type	memcpy, @function
.global	memcpy
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memmem.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memmem.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+16] = length of needle
					# [esp+12] = address of needle
					# [esp+8] = length of haystack
					# [esp+4] = address of haystack
memmem:
	xor	eax, eax		# eax = address of needle in haystack = 0
	mov	ecx, [esp+8]		# ecx = length of haystack
	test	ecx, ecx
	jz	.Lempty			# length of haystack = 0?

	mov	edx, [esp+16]		# edx = length of needle
	cmp	edx, ecx
	ja	.Lempty			# length of needle > length of haystack?

	mov	eax, [esp+4]		# eax = address of haystack
	test	edx, edx
	jz	.Lempty			# length of needle = 0?

	push	ebx
	push	edi
	mov	edi, eax		# edi = address of haystack
	push	esi
.Lsearch:
	mov	esi, [esp+24]		# esi = address of needle
	mov	al, [esi]		# al = first character of needle
					# edi = current address in haystack
	repne	scasb			# edi = next address in haystack,
					# ecx = current length of haystack
	jne	.Lbreak			# (first character of) needle not in haystack?

	dec	ecx			# ecx = next length of haystack

	mov	al, [esi+edx-1]		# al = last character of needle
	cmp	al, [edi+edx-2]
	jne	.Lcontinue		# last character of needle not in haystack?
.Lcompare:
	mov	eax, edi		# eax = next address in haystack
	mov	ebx, ecx		# ebx = next length of haystack
.if 0
	dec	edi			# edi = address of matching character
					#     = current address in haystack
					# esi = address of needle
	mov	ecx, edx		# ecx = length of needle
.else
					# edi = next address in haystack
	inc	esi			# esi = address of needle + 1
	mov	ecx, edx
	dec	ecx			# ecx = length of needle - 1
#	jz	.Lmatch			# needle in haystack?
.endif
	repe	cmpsb
	je	.Lmatch			# needle in haystack?

	mov	edi, eax		# edi = current address in haystack
	mov	ecx, ebx		# ecx = current length of haystack
.Lcontinue:
	cmp	ecx, edx
	jae	.Lsearch		# length of haystack >= length of needle?
.Lbreak:
	xor	eax, eax
	pop	esi
	pop	edi
	pop	ebx
.Lempty:
	ret
.Lmatch:
	dec	eax			# eax = address of needle in haystack
	pop	esi
	pop	edi
	pop	ebx
	ret

.size	memmem, .-memmem
.type	memmem, @function
.global	memmem
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memmove.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memmove.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = address of source
					# [esp+4] = address of destination
memmove:
	mov	eax, [esp+4]		# eax = address of destination
	mov	edx, [esp+8]		# edx = address of source
	cmp	edx, eax
	je	.Lnone			# address of source = address of destination?

	mov	ecx, [esp+12]		# ecx = count
#	jecxz	.Lnone			# count = 0?

	xchg	esi, edx		# esi = address of source
	xchg	edi, eax		# edi = address of destination
	ja	.Ldefault		# address of source > address of destination?
.Loverlap:
	lea	edi, [edi+ecx-1]
	lea	esi, [esi+ecx-1]
	std
.Ldefault:
	rep	movsb
	cld
	mov	esi, edx
	mov	edi, eax
	mov	eax, [esp+4]		# eax = address of destination
.Lnone:
	ret

.size	memmove, .-memmove
.type	memmove, @function
.global	memmove
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memrchr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = character
					# [esp+4] = address of buffer
memrchr:
	xor	eax, eax		# eax = 0
	cmp	eax, [esp+12]
	je	.Lreturn		# count = 0?

	cdq				# edx = 0
	mov	dl, [esp+8]		# edx = character
	imul	edx, 0x01010101		# edx = character
					#     | character << 8
					#     | character << 16
					#     | character << 24
	mov	ecx, [esp+4]		# ecx = address of buffer
	add	ecx, [esp+12]		# ecx = address after buffer
	push	ebx
	mov	ebx, ecx		# ebx = address after buffer
	and	ecx, 3			# ecx = address after buffer % 4
					#     = number of tail bytes
	jz	.Laligned		# address after buffer % 4 = 0?
.Lunaligned:
	sub	ebx, ecx		# ebx = aligned address of tail bytes
	shl	ecx, 3			# ecx = number of tail bytes * 8
					#     = number of tail bits
	dec	eax			# eax = ~0
	shl	eax, cl			# eax = 0 for tail bytes, ~0 elsewhere
	mov	ecx, [ebx]		# ecx = tail bytes
	xor	ecx, edx		# ecx = tail bytes ^ multiplied character
	or	eax, ecx		# eax = '\0' for matching tail bytes
	jmp	.Ltwiddle
.Lprevious:
	sub	ebx, 4			# ebx = address of previous 4 aligned bytes
	cmp	ebx, [esp+8]
	jb	.Lnull			# address before buffer?
.Laligned:
	mov	eax, [ebx]		# eax = previous 4 aligned bytes
	xor	eax, edx		# eax = previous 4 aligned bytes ^ multiplied character
					#     = '\0' for matching bytes
.Ltwiddle:
	mov	ecx, eax
	sub	eax, 0x01010101
	not	ecx
	and	eax, 0x80808080
	and	eax, ecx		# eax = '\200' for matching bytes, '\0' elsewhere
	jz	.Lprevious		# no match in any byte?
.Lmatch:
	bsr	eax, eax		# eax = offset of matching byte * 8 + 7
					#     = {31, 23, 15, 7}
	shr	eax, 3			# eax = offset of matching byte
					#     = {3, 2, 1, 0}
	add	eax, ebx		# eax = address of matching byte
	cmp	eax, [esp+8]		# CF = (address of matching byte < address of buffer)
	cmc				# CF = (address of matching byte >= address of buffer)
	sbb	ecx, ecx		# ecx = (address inside buffer) ? -1 : 0
	and	eax, ecx		# eax = address of character
.Lnull:
	pop	ebx
.Lreturn:
	ret

.size	memrchr, .-memrchr
.type	memrchr, @function
.global	memrchr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memrchr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = character
					# [esp+4] = address of buffer
memrchr:
	mov	ecx, [esp+12]		# ecx = count
	test	ecx, ecx
	jz	.Lnone			# count = 0?

	mov	edx, [esp+8]		# dl = character
	mov	eax, [esp+4]		# eax = address of buffer
	lea	eax, [eax+ecx-1]	# eax = end of buffer
.Lloop:
	cmp	dl, [eax]
	je	.Lmatch			# dl = *buffer?

	dec	eax			# eax = --buffer
	dec	ecx			# ecx = --count
	jnz	.Lloop			# count <> 0?
.Lnone:
	mov	eax, ecx
.Lmatch:
	ret

.size	memrchr, .-memrchr
.type	memrchr, @function
.global	memrchr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memrchr.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memrchr.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = character
					# [esp+4] = address of buffer
memrchr:
	mov	edx, edi
	mov	edi, [esp+4]		# edi = address of buffer
	mov	eax, [esp+8]		# eax = character
	mov	ecx, [esp+12]		# ecx = count
	lea	edi, [edi+ecx-1]	# edi = end of buffer
	std
	repne	scasb
	cld
	inc	edi			# edi = address of character
	neg	ecx			# CF = (ecx <> 0)
					#    = ([edi] = character)
	sbb	eax, eax		# eax = (ecx = 0) ? 0 : -1
	and	eax, edi		# eax = (ecx = 0) ? 0 : address of character
	mov	edi, edx
	ret

.size	memrchr, .-memrchr
.type	memrchr, @function
.global	memrchr
.end
+EOF

as --32 --no-pad-sections --noexecstack -o memset.o <<+\E\O\F
# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

.file	"memset.s"
.arch	generic32
.code32
.intel_syntax noprefix
.text
					# [esp+12] = count
					# [esp+8] = character
					# [esp+4] = address of buffer
memset:
	mov	edx, edi
	mov	edi, [esp+4]		# edi = address of buffer
	mov	eax, [esp+8]		# eax = character
	mov	ecx, [esp+12]		# ecx = count
#	jecxz	.Lnone			# count = 0?

	rep	stosb
.Lnone:
	mov	eax, [esp+4]		# eax = address of buffer
	mov	edi, edx
	ret

.size	memset, .-memset
.type	memset, @function
.global	memset
.end
+EOF

ar rsv memory-i386.a memory.o memchr.o memcmp.o memcpy.o memmem.o memmove.o memrchr.o memset.o
rm memory.o memchr.o memcmp.o memcpy.o memmem.o memmove.o memrchr.o memset.o
else
echo "Usage: $0 { amd64 | i386 }"
echo
echo "Create object library 'memory-amd64.a' or 'memory-i386.a' with properly"
echo "optimised memory routines for Clang and GCC, as (partial) replacement for"
echo "either 'libclang-rt.builtins-x86_64.a' or 'libclang-rt.builtins-i386.a' as"
echo "well as 'libgcc.a'"
fi
